"""
Async client for the Overpass API.

[Release Notes](https://github.com/timwie/aio-overpass/blob/main/RELEASES.md)

[Examples](https://github.com/timwie/aio-overpass/tree/main/examples)
"""

import importlib.metadata
from pathlib import Path


__version__: str = importlib.metadata.version("aio-overpass")

# we add this to all modules for pdoc;
# see https://pdoc.dev/docs/pdoc.html#use-numpydoc-or-google-docstrings
__docformat__ = "google"

# we also use __all__ in all modules for pdoc; this lets us control the order
__all__ = (
    "__version__",
    "Client",
    "ClientError",
    "Query",
    "client",
    "element",
    "error",
    "pt",
    "pt_ordered",
    "ql",
    "query",
    "spatial",
)

from .client import Client
from .error import ClientError
from .query import Query


# extend the module's docstring
for filename in ("usage.md", "extras.md", "coordinates.md"):
    __doc__ += "\n<br>\n"
    __doc__ += (Path(__file__).parent / "doc" / filename).read_text(encoding="utf-8")



"""
Error types.

```
                            (ClientError)
                                  ╷
                   ┌──────────────┼────────────┬────────────────┬────────────────┐
                   ╵              ╵            ╵                ╵                ╵
               RunnerError   (QueryError)   GiveupError    ResponseError     CallError
                                  ╷                             ╷                ╷
         ┌────────────────────────┼──────────────────────┐      │                │
         ╵                        ╵                      ╵      ╵                ╵
QueryLanguageError         QueryRejectError          QueryResponseError   CallTimeoutError
```
"""

import asyncio
import html
import logging
import math
import re
from dataclasses import dataclass
from enum import Enum, auto
from json import JSONDecodeError
from typing import NoReturn, TypeAlias, TypeGuard

import aiohttp
import aiohttp.typedefs


__docformat__ = "google"
__all__ = (
    "AlreadyRunningError",
    "ClientError",
    "CallError",
    "CallTimeoutError",
    "ResponseError",
    "ResponseErrorCause",
    "GiveupError",
    "GiveupCause",
    "QueryError",
    "QueryLanguageError",
    "QueryRejectError",
    "QueryRejectCause",
    "QueryResponseError",
    "RunnerError",
    "is_call_err",
    "is_call_timeout",
    "is_exceeding_maxsize",
    "is_exceeding_timeout",
    "is_gateway_rejection",
    "is_giveup_err",
    "is_rejection",
    "is_runtime_rejection",
    "is_server_error",
    "is_too_busy",
    "is_too_many_queries",
)


@dataclass(kw_only=True)
class AlreadyRunningError(ValueError):
    """
    Raised when trying to run a query while it's already being run.

    Attributes:
        kwargs: the query's ``kwargs``
    """

    kwargs: dict

    def __str__(self) -> str:
        return f"query{self.kwargs!r} is already being run"


class ClientError(Exception):
    """Base exception for failed Overpass API requests and queries."""

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return False


@dataclass(kw_only=True)
class RunnerError(ClientError):
    """
    The query runner raised an exception.

    This is an unexpected error, in contrast to the runner intentionally
    raising ``query.error`` to stop retrying. The cause of this error
    is therefore anything but a ``ClientError``.

    Attributes:
        cause: the exception that caused this error
    """

    cause: BaseException

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return False

    def __str__(self) -> str:
        return str(self.cause)


@dataclass(kw_only=True)
class CallError(ClientError):
    """
    Failed to make an API request.

    This error is raised when the client failed to get any response,
    f.e. due to connection issues.

    Attributes:
        cause: the exception that caused this error
    """

    cause: aiohttp.ClientError | asyncio.TimeoutError

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return True

    def __str__(self) -> str:
        return str(self.cause)


@dataclass(kw_only=True)
class CallTimeoutError(CallError):
    """
    An API request timed out.

    Attributes:
        cause: the exception that caused this error
        after_secs: the configured timeout for the request
    """

    cause: asyncio.TimeoutError
    after_secs: float

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return True

    def __str__(self) -> str:
        return str(self.cause)


ResponseErrorCause: TypeAlias = aiohttp.ClientResponseError | JSONDecodeError | ValueError
"""Causes for a ``ResponseError``."""


@dataclass(kw_only=True)
class ResponseError(ClientError):
    """
    Unexpected API response.

    On the one hand, this can be an error that happened on the Overpass API instance,
    which is usually signalled by a status code ``>= 500``. In rare cases,
    the ``cause`` being a ``JSONDecodeError`` also signals this, since it can happen
    that the API returns a cutoff JSON response. In both of these cases,
    ``is_server_error()`` will return ``True``.

    On the other hand, this error may indicate that a request failed, but we can't
    specifically say why. This could be a bug on our end, since the client
    is meant to process almost any response of an Overpass API instance.
    Here, ``is_server_error()`` will return ``False``, and the default query runner
    will log the response body.

    Attributes:
        response: the unexpected response
        body: the response body
        cause: an optional exception that may have caused this error
    """

    response: aiohttp.ClientResponse
    body: str
    cause: ResponseErrorCause | None

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return True

    @property
    def is_server_error(self) -> bool:
        """Returns ``True`` if this presumably a server-side error."""
        # see class doc for details
        return self.response.status >= 500 or isinstance(self.cause, JSONDecodeError)

    def __str__(self) -> str:
        if self.cause is None:
            return f"unexpected response ({self.response.status})"
        return f"unexpected response ({self.response.status}): {self.cause}"


class GiveupCause(Enum):
    """Details on why a query was given up on."""

    RUN_TIMEOUT_BEFORE_STATUS_CALL = auto()
    """Query's ``run_timeout_secs`` elapsed before a status call before the actual query call."""

    RUN_TIMEOUT_BY_COOLDOWN = auto()
    """Query received a timeout that would exceed its remaining ``run_timeout_secs``."""

    RUN_TIMEOUT_BEFORE_QUERY_CALL = auto()
    """Query's ``run_timeout_secs`` elapsed before the query call."""

    EXPECTING_QUERY_TIMEOUT = auto()
    """
    The query's limited [timeout:*] setting is likely too low.

    A query's [timeout:*] setting might be lowered by the client to the remainder of
    ``run_timeout_secs``. If that happens, and that duration is also lower than a duration
    after which it was cancelled at the server previously, then this is the cause to give
    up on the query.

    See Also:
        ``QueryRejectError.timed_out_after_secs``
    """

    RUN_TIMEOUT_DURING_QUERY_CALL = auto()
    """Query's ``run_timeout_secs`` elapsed while making the query call, before any response."""


@dataclass(kw_only=True)
class GiveupError(ClientError):
    """
    The client spent or would spend too long running a query, and gave up.

    This error is raised when the run timeout duration set by a query runner
    is or would be exceeded.

    Attributes:
        kwargs: the query's ``kwargs``
        after_secs: the total time spent on the query
        cause: why the query was given up on
    """

    kwargs: dict
    after_secs: float
    cause: GiveupCause

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return False

    def __str__(self) -> str:
        return f"gave up on query{self.kwargs!r} after {self.after_secs:.01f} seconds"


@dataclass(kw_only=True)
class QueryError(ClientError):
    """
    Base exception for queries that failed at the Overpass API server.

    Attributes:
        kwargs: the query's ``kwargs``
        remarks: the error remarks provided by the API
    """

    kwargs: dict
    remarks: list[str]

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return False

    def __str__(self) -> str:
        first = f"'{self.remarks[0]}'"
        rest = f" (+{len(self.remarks) - 1} more)" if len(self.remarks) > 1 else ""
        return f"query{self.kwargs!r} failed: {first}{rest}"


@dataclass(kw_only=True)
class QueryResponseError(ResponseError, QueryError):
    """
    Unexpected query response.

    This error is raised when a query fails (thus extends ``QueryError``),
    but we can't specifically say why (thus also extends ``ResponseError``).
    """

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return ResponseError.should_retry.fget(self)  # type: ignore[attr-defined]

    def __str__(self) -> str:
        if self.remarks:
            first = f"'{self.remarks[0]}'"
            rest = f" (+{len(self.remarks) - 1} more)" if len(self.remarks) > 1 else ""
            return f"query{self.kwargs!r} failed: {first}{rest}"

        return f"query{self.kwargs!r} failed with status {self.response.status}"


@dataclass(kw_only=True)
class QueryLanguageError(QueryError):
    """
    Indicates the query's QL code is not valid.

    Retrying is pointless when encountering this error.
    """

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return False


class QueryRejectCause(Enum):
    """Details why a query was rejected or cancelled by an API server."""

    TOO_BUSY = auto()
    """
    Gateway rejection. The server has already so much load that the request cannot be executed.

    Smaller ``[timeout:*]`` and/or ``[maxsize:*]`` values might make the request acceptable.
    """

    TOO_MANY_QUERIES = auto()
    """
    Gateway rejection. There are no open slots for queries from your IP address.

    Running queries take up a slot, and the number of slots is limited. A client
    will only run as many concurrent requests as there are slots, which should make this
    a rare error, assuming you are not making requests through another client.
    """

    EXCEEDED_TIMEOUT = auto()
    """
    Runtime rejection. The query has been (or surely will be) running longer than its proposed
    ``[timeout:*]``, and has been cancelled by the server.

    A higher ``[timeout:*]`` value might allow the query run to completion, but also makes it
    more likely to be rejected by a server under heavy load, before executing it (see ``TOO_BUSY``).
    """

    EXCEEDED_MAXSIZE = auto()
    """
    Runtime rejection. The memory required to execute the query has (or surely will) exceed
    its proposed ``[maxsize:*]``, and has been cancelled by the server.

    A higher ``[maxsize:*]`` value might allow the query run to completion, but also makes it
    more likely to be rejected by a server under heavy load, before executing it (see ``TOO_BUSY``).
    """

    def __str__(self) -> str:
        match self:
            case QueryRejectCause.TOO_BUSY:
                return "server too busy"
            case QueryRejectCause.TOO_MANY_QUERIES:
                return "too many queries"
            case QueryRejectCause.EXCEEDED_TIMEOUT:
                return "exceeded 'timeout'"
            case QueryRejectCause.EXCEEDED_MAXSIZE:
                return "exceeded 'maxsize'"
            case _:
                raise AssertionError


@dataclass(kw_only=True)
class QueryRejectError(QueryError):
    """
    A query was rejected or cancelled by the API server.

    Attributes:
        cause: why the query was rejected or cancelled
        timed_out_after_secs: if ``cause`` is ``EXCEEDED_TIMEOUT``, this is the amount
                              of seconds after which the query timed out
        oom_using_mib: if ```cause`` is ``EXCEEDED_MAXSIZE``, this is the amount of RAM
                       the query used before it ran out
    """

    cause: QueryRejectCause
    timed_out_after_secs: int | None
    oom_using_mib: int | None

    @property
    def should_retry(self) -> bool:
        """Returns ``True`` if it's worth retrying when encountering this error."""
        return True

    def __str__(self) -> str:
        match self.cause:
            case QueryRejectCause.TOO_BUSY | QueryRejectCause.TOO_MANY_QUERIES:
                rejection = "rejected"
            case QueryRejectCause.EXCEEDED_TIMEOUT:
                rejection = f"cancelled after {self.timed_out_after_secs}s"
            case QueryRejectCause.EXCEEDED_MAXSIZE:
                rejection = f"cancelled using {self.oom_using_mib}"
            case _:
                raise AssertionError(self.cause)

        return f"{rejection}: {self.cause}"


async def _raise_for_request_error(err: aiohttp.ClientError) -> NoReturn:
    """
    Raise an exception caused by the given request error.

    Raises:
        - ``CallError`` if ``obj`` is an ``aiohttp.ClientError``,
          but not an ``aiohttp.ClientResponseError``.
        - ``ResponseError`` otherwise.
    """
    if isinstance(err, aiohttp.ClientResponseError):
        response = err.history[-1]
        await _raise_for_response(response, err)

    raise CallError(cause=err) from err


async def _raise_for_response(
    response: aiohttp.ClientResponse,
    cause: ResponseErrorCause | None,
) -> NoReturn:
    """Raise a ``ResponseError`` with an optional cause."""
    err = ResponseError(
        response=response,
        body=await response.text(),
        cause=cause,
    )
    if cause:
        raise err from cause
    raise err


async def _result_or_raise(
    response: aiohttp.ClientResponse,
    query_kwargs: dict,
    query_logger: logging.Logger,
) -> dict:
    """
    Try to extract the query result set from a response.

    Raises:
        CallError: When there is any sort of connection error.
        RejectError: When encountering "Too Many Requests" or "Gateway Timeout";
                     when there's a JSON remark indicating query rejection or cancellation;
                     when there's an HTML error message indicating query rejection or cancellation.
        QueryError: When there's any other JSON remark or HTML error message.
        ResponseError: When encountering an unexpected response.
    """
    await __raise_for_plaintext_result(response)

    await __raise_for_html_result(response, query_kwargs, query_logger)

    return await __raise_for_json_result(response, query_kwargs, query_logger)


async def __raise_for_json_result(
    response: aiohttp.ClientResponse,
    query_kwargs: dict,
    query_logger: logging.Logger,
) -> dict:
    try:
        json = await response.json()
        if json is None:
            await _raise_for_response(response, cause=None)
    except aiohttp.ClientResponseError as err:
        await _raise_for_response(response, cause=err)
    except JSONDecodeError as err:
        await _raise_for_response(response, cause=err)

    if remark := json.get("remark"):
        if timeout_cause := __match_reject_cause(remark, query_logger):
            raise QueryRejectError(
                kwargs=query_kwargs,
                remarks=[remark],
                cause=timeout_cause,
                oom_using_mib=__match_oom_after(remark),
                timed_out_after_secs=__match_timeout_after(remark),
            )

        raise QueryResponseError(
            response=response,
            body=await response.text(),
            cause=None,
            kwargs=query_kwargs,
            remarks=[remark],
        )

    expected_fields = ("version", "generator", "osm3s", "elements")
    expected_osm3s_fields = ("timestamp_osm_base", "copyright")
    if any(f not in json for f in expected_fields) or any(
        f not in json["osm3s"] for f in expected_osm3s_fields
    ):
        await _raise_for_response(response, cause=None)

    return json


async def __raise_for_html_result(
    response: aiohttp.ClientResponse,
    query_kwargs: dict,
    query_logger: logging.Logger,
) -> None:
    """
    Raise a fitting exception based on error remarks in an HTML response.

    Raises:
        RejectError: if one of the error remarks indicate that the query was rejected or cancelled
        QueryError: when encountering other error remarks
        ResponseError: when error remarks cannot be extracted from the response
    """
    if response.content_type != "text/html":
        return

    text = await response.text()

    pattern = re.compile("Error</strong>: (.+?)</p>", re.DOTALL)
    errors = [html.unescape(err.strip()) for err in pattern.findall(text)]

    if not errors:  # unexpected format
        await _raise_for_response(response, cause=None)

    if any(__is_ql_error(msg, query_logger) for msg in errors):
        raise QueryLanguageError(kwargs=query_kwargs, remarks=errors)

    reject_causes = [cause for err in errors if (cause := __match_reject_cause(err, query_logger))]
    oom_using_mib = next((mib for err in errors if (mib := __match_oom_after(err))), None)
    timed_out_after_secs = next(
        (secs for err in errors if (secs := __match_timeout_after(err))), None
    )

    if reject_causes:
        raise QueryRejectError(
            kwargs=query_kwargs,
            remarks=errors,
            cause=reject_causes[0],
            oom_using_mib=oom_using_mib,
            timed_out_after_secs=timed_out_after_secs,
        )

    raise QueryResponseError(
        response=response,
        body=await response.text(),
        cause=None,
        kwargs=query_kwargs,
        remarks=errors,
    )


async def __raise_for_plaintext_result(response: aiohttp.ClientResponse) -> None:
    if response.content_type != "text/plain":
        return
    await _raise_for_response(response, cause=None)


def __match_reject_cause(error_msg: str, query_logger: logging.Logger) -> QueryRejectCause | None:
    """
    Check if the given error message indicates that a query was rejected or cancelled.

    AFAIK, neither the 'remarks' in JSON responses, nor the errors listed in HTML responses
    are neatly listed somewhere, but it seems matching a small subset of remarks is enough
    to identify recoverable errors.

    References:
        - Related: https://github.com/DinoTools/python-overpy/issues/62
        - Examples in the API source: https://github.com/drolbr/Overpass-API/search?q=runtime_error
    """
    if "Please check /api/status for the quota of your IP address" in error_msg:
        cause = QueryRejectCause.TOO_MANY_QUERIES

    elif "The server is probably too busy to handle your request" in error_msg:
        cause = QueryRejectCause.TOO_BUSY

    elif "Query timed out" in error_msg:
        cause = QueryRejectCause.EXCEEDED_TIMEOUT

    elif "out of memory" in error_msg:
        cause = QueryRejectCause.EXCEEDED_MAXSIZE

    else:
        cause_cls = QueryRejectCause.__class__.__name__
        query_logger.debug(f"does not match any {cause_cls}: {error_msg!r}")
        return None

    query_logger.debug(f"matches {cause}: {error_msg!r}")
    return cause


def __match_oom_after(error_msg: str) -> int | None:
    pattern = re.compile(
        r"^runtime error: Query run out of memory in \".+\""
        r" at line \d+ using about (\d+) MB of RAM\.$"
    )
    if m := pattern.match(error_msg):
        mb = int(m.group(1))
        return math.ceil((mb * 1000**2) / 1024**2)
    return None


def __match_timeout_after(error_msg: str) -> int | None:
    pattern = re.compile(
        r"^runtime error: Query timed out in \".+\" at line \d+ after (\d+) seconds\.$"
    )
    if m := pattern.match(error_msg):
        return int(m.group(1))
    return None


def __is_ql_error(error_msg: str, query_logger: logging.Logger) -> bool:
    """Check if the given error message indicates that a query has bad QL code."""
    is_ql_error = (
        "encoding error:" in error_msg
        or "parse error:" in error_msg
        or "static error:" in error_msg
    )

    if is_ql_error:
        query_logger.debug(f"is a QL error: {error_msg!r}")
        return True

    query_logger.debug(f"not a QL error: {error_msg!r}")
    return False


def is_call_err(err: ClientError | None) -> TypeGuard[CallError]:
    """``True`` if this is a ``CallError``."""
    return isinstance(err, CallError)


def is_call_timeout(err: ClientError | None) -> TypeGuard[CallTimeoutError]:
    """``True`` if this is a ``CallTimeoutError``."""
    return isinstance(err, CallTimeoutError)


def is_giveup_err(err: ClientError | None) -> TypeGuard[GiveupError]:
    """``True`` if this is a ``GiveupError``."""
    return isinstance(err, GiveupError)


def is_server_error(err: ClientError | None) -> TypeGuard[ResponseError]:
    """``True`` if this is a ``ResponseError`` presumably cause by a server-side error."""
    return isinstance(err, ResponseError) and err.is_server_error


def is_rejection(err: ClientError | None) -> TypeGuard[QueryRejectError]:
    """``True`` if this is a ``QueryRejectError``."""
    return isinstance(err, QueryRejectError)


def is_gateway_rejection(err: ClientError | None) -> TypeGuard[QueryRejectError]:
    """``True`` if this is a ``QueryRejectError`` with gateway rejection."""
    return isinstance(err, QueryRejectError) and err.cause in {
        QueryRejectCause.TOO_MANY_QUERIES,
        QueryRejectCause.TOO_BUSY,
    }


def is_too_many_queries(err: ClientError | None) -> TypeGuard[QueryRejectError]:
    """``True`` if this is a ``QueryRejectError`` with cause ``TOO_MANY_QUERIES``."""
    return isinstance(err, QueryRejectError) and err.cause is QueryRejectCause.TOO_MANY_QUERIES


def is_too_busy(err: ClientError | None) -> TypeGuard[QueryRejectError]:
    """``True`` if this is a ``QueryRejectError`` with cause ``TOO_BUSY``."""
    return isinstance(err, QueryRejectError) and err.cause is QueryRejectCause.TOO_BUSY


def is_runtime_rejection(err: ClientError | None) -> TypeGuard[QueryRejectError]:
    """``True`` if this is a ``QueryRejectError`` with runtime rejection."""
    return isinstance(err, QueryRejectError) and err.cause in {
        QueryRejectCause.EXCEEDED_MAXSIZE,
        QueryRejectCause.EXCEEDED_TIMEOUT,
    }


def is_exceeding_maxsize(err: ClientError | None) -> TypeGuard[QueryRejectError]:
    """``True`` if this is a ``GiveupError`` with cause ``EXCEEDED_MAXSIZE``."""
    return isinstance(err, QueryRejectError) and err.cause is QueryRejectCause.EXCEEDED_MAXSIZE


def is_exceeding_timeout(err: ClientError | None) -> TypeGuard[QueryRejectError]:
    """``True`` if this is a ``GiveupError`` with cause ``EXCEEDED_TIMEOUT``."""
    return isinstance(err, QueryRejectError) and err.cause is QueryRejectCause.EXCEEDED_TIMEOUT



import os
from typing import Final


__docformat__ = "google"
__all__ = (
    "IS_CI",
    "IS_UNIT_TEST",
    "FORCE_DISABLE_CACHE",
)

IS_CI: Final[bool] = "GITHUB_ACTIONS" in os.environ
IS_UNIT_TEST: Final[bool] = "PYTEST_VERSION" in os.environ
FORCE_DISABLE_CACHE: Final[bool] = IS_CI and not IS_UNIT_TEST



"""Basic definitions for (groups of) geospatial objects."""

from abc import ABC, abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any, TypeAlias


__docformat__ = "google"
__all__ = (
    "GeoJsonDict",
    "SpatialDict",
    "Spatial",
)


GeoJsonDict: TypeAlias = dict[str, Any]
"""A dictionary representing a GeoJSON object."""


@dataclass(kw_only=True, slots=True)
class SpatialDict:
    """
    Mapping of spatial objects with the ``__geo_interface__`` property.

    Objects of this class have the ``__geo_interface__`` property following a protocol
    [proposed](https://gist.github.com/sgillies/2217756) by Sean Gillies, which can make
    it easier to use spatial data in other Python software. An example of this is the ``shape()``
    function that builds Shapely geometries from any object with the ``__geo_interface__`` property.

    Attributes:
        __geo_interface__: this is the proposed property that contains the spatial data
    """

    __geo_interface__: dict


class Spatial(ABC):
    """
    Base class for (groups of) geospatial objects.

    Classes that represent spatial features extend this class and implement the
    ``geojson`` property. Exporting objects in the GeoJSON format should make it possible
    to integrate them with other tools for visualization, or further analysis.
    The ability to re-import the exported GeoJSON structures as ``Spatial`` objects is not
    considered here.
    """

    __slots__ = ("__validated__",)  # we use that field in tests

    @property
    @abstractmethod
    def geojson(self) -> GeoJsonDict:
        """
        A mapping of this object, using the GeoJSON format.

        The coordinate reference system for all GeoJSON coordinates is ``CRS:84``,
        which means every coordinate is a tuple of longitude and latitude (in that order)
        on the WGS 84 ellipsoid. Note that this order is flipped for all Shapely geometries
        that represent OpenStreetMap elements (latitude first, then longitude).

        References:
            - https://osmdata.openstreetmap.de/info/projections.html
            - https://tools.ietf.org/html/rfc7946#section-4
        """
        raise NotImplementedError

    @property
    def geo_interfaces(self) -> Iterator[SpatialDict]:
        """A mapping of this object to ``SpatialDict``s that implement ``__geo_interface__``."""
        geojson = self.geojson
        match geojson["type"]:
            case "FeatureCollection":
                for feature in geojson["features"]:
                    yield SpatialDict(__geo_interface__=feature)
            case _:
                yield SpatialDict(__geo_interface__=geojson)



"""Interface for making API calls."""

import math
import re
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager, suppress
from dataclasses import dataclass
from typing import Final
from urllib.parse import urljoin

from aio_overpass import __version__
from aio_overpass._clock import sleep
from aio_overpass.error import (
    AlreadyRunningError,
    CallError,
    CallTimeoutError,
    ClientError,
    GiveupCause,
    GiveupError,
    RunnerError,
    _raise_for_request_error,
    _raise_for_response,
    _result_or_raise,
    is_too_many_queries,
)
from aio_overpass.query import DefaultQueryRunner, Query, QueryRunner

import aiohttp
from aiohttp import ClientTimeout


__docformat__ = "google"
__all__ = (
    "Client",
    "Status",
    "DEFAULT_INSTANCE",
    "DEFAULT_USER_AGENT",
)


DEFAULT_INSTANCE: Final[str] = "https://overpass-api.de/api/"
"""Main Overpass API instance."""

DEFAULT_USER_AGENT: Final[str] = (
    f"aio-overpass/{__version__} (https://github.com/timwie/aio-overpass)"
)
"""User agent that points to the ``aio-overpass`` repo."""


@dataclass(kw_only=True, slots=True)
class Status:
    """
    Information about the API server's rate limit.

    Attributes:
        slots: The maximum number of concurrent queries per IP
               (or ``None`` if there is no rate limit).
        free_slots: The number of slots open for this IP
                    (or ``None`` if there is no rate limit).
        cooldown_secs: The number of seconds until a slot opens for this IP
                       (or 0 if there is a free slot).
        endpoint: Announced endpoint. For example, there are two distinct servers
                  that both can be reached by the main Overpass API instance.
                  Depending on server load, a query may be sent to either of them.
                  This value is the server name, f.e. ``"gall.openstreetmap.de/"``.
        nb_running_queries: Number of currently running queries for this IP.
    """

    slots: int | None
    free_slots: int | None
    cooldown_secs: int
    endpoint: str | None
    nb_running_queries: int

    def __repr__(self) -> str:
        rep = f"{type(self).__name__}(free_slots="
        rep += f"{self.free_slots}/{self.slots}" if self.slots else "∞"
        rep += f", cooldown={self.cooldown_secs}s"
        rep += f", running={self.nb_running_queries}"
        if self.endpoint:
            rep += f", endpoint={self.endpoint!r}"
        rep += ")"
        return rep


class Client:
    """
    A client for the Overpass API.

    Requests are rate-limited according to the configured number of slots per IP for the specified
    API server. By default, queries are retried whenever the server is too busy, or the rate limit
    was exceeded. Custom query runners can be used to implement your own retry strategy.

    Args:
        url: The url of an Overpass API instance. Defaults to the main Overpass API instance.
        user_agent: A string used for the User-Agent header. It is good practice to provide a string
                    that identifies your application, and includes a way to contact you (f.e. an
                    e-mail, or a link to a repository). This is important if you make too many
                    requests, or queries that require a lot of resources.
        concurrency: The maximum number of simultaneous connections. In practice the amount
                     of concurrent queries may be limited by the number of slots it provides for
                     each IP.
        status_timeout_secs: If set, status requests to the Overpass API will time out after
                             this duration in seconds. Defaults to no timeout.
        runner: You can provide another query runner if you want to implement your own retry
                strategy.

    References:
        - https://wiki.openstreetmap.org/wiki/Overpass_API#Public_Overpass_API_instances
    """

    __slots__ = (
        "_concurrency",
        "_maybe_session",
        "_runner",
        "_status_timeout_secs",
        "_url",
        "_user_agent",
    )

    def __init__(
        self,
        url: str = DEFAULT_INSTANCE,
        user_agent: str = DEFAULT_USER_AGENT,
        concurrency: int = 32,
        status_timeout_secs: float | None = None,
        runner: QueryRunner | None = None,
    ) -> None:
        if concurrency <= 0:
            msg = "'concurrency' must be > 0"
            raise ValueError(msg)
        if status_timeout_secs is not None and (
            not math.isfinite(status_timeout_secs) or status_timeout_secs <= 0.0
        ):
            msg = "'status_timeout_secs' must be finite > 0"
            raise ValueError(msg)

        self._url: Final[str] = url
        self._user_agent: Final[str] = user_agent
        self._concurrency: Final[int] = concurrency
        self._status_timeout_secs: Final[float | None] = status_timeout_secs
        self._runner: Final[QueryRunner] = runner or DefaultQueryRunner()

        self._maybe_session: aiohttp.ClientSession | None = None

    def _session(self) -> aiohttp.ClientSession:
        """The session used for all requests of this client."""
        if not self._maybe_session or self._maybe_session.closed:
            headers = {"User-Agent": self._user_agent}
            connector = aiohttp.TCPConnector(limit=self._concurrency)
            self._maybe_session = aiohttp.ClientSession(headers=headers, connector=connector)

        return self._maybe_session

    async def close(self) -> None:
        """Cancel all running queries and close the underlying session."""
        if self._maybe_session and not self._maybe_session.closed:
            # do not care if this fails
            with suppress(CallError):
                _ = await self.cancel_queries()

            # is raised when there are still active queries. that's ok
            with suppress(aiohttp.ServerDisconnectedError):
                await self._maybe_session.close()

    async def _status(self, timeout: ClientTimeout | None = None) -> "Status":
        endpoint = urljoin(self._url, "status")
        timeout = timeout or aiohttp.ClientTimeout(total=self._status_timeout_secs)
        async with (
            _map_request_error(timeout),
            self._session().get(url=endpoint, timeout=timeout) as response,
        ):
            return await _parse_status(response)

    async def status(self) -> Status:
        """
        Check the current API status.

        The timeout of this request is configured with the ``status_timeout_secs`` argument.

        Raises:
            ClientError: if the status could not be looked up
        """
        return await self._status()

    async def cancel_queries(self, timeout_secs: float | None = None) -> int:
        """
        Cancel all running queries.

        This can be used to terminate runaway queries that prevent you from sending new ones.

        Returns:
            the number of terminated queries

        Raises:
            ClientError: if the request to cancel queries failed
        """
        if timeout_secs is not None and (not math.isfinite(timeout_secs) or timeout_secs <= 0.0):
            msg = "'timeout_secs' must be finite > 0"
            raise ValueError(msg)

        timeout = aiohttp.ClientTimeout(total=timeout_secs) if timeout_secs else None
        headers = {"User-Agent": self._user_agent}
        endpoint = urljoin(self._url, "kill_my_queries")

        # use a new session here to get around our concurrency limit
        async with (
            aiohttp.ClientSession(headers=headers) as session,
            _map_request_error(timeout),
            session.get(endpoint, timeout=timeout) as response,
        ):
            body = await response.text()
            killed_pids = re.findall("\\(pid (\\d+)\\)", body)
            return len(set(killed_pids))

    async def run_query(self, query: Query, *, raise_on_failure: bool = True) -> None:
        """
        Send a query to the API, and await its completion.

        "Running" the query entails acquiring a connection from the pool, the query requests
        themselves (which may be retried), status requests when the server is busy,
        and cooldown periods.

        The query runner is invoked before every try, and once after the last try.

        To run multiple queries concurrently, wrap the returned coroutines in an ``asyncio`` task,
        f.e. with ``asyncio.create_task()`` and subsequent ``asyncio.gather()``.

        Args:
            query: the query to run on this API instance
            raise_on_failure: if ``True``, raises ``query.error`` if the query failed

        Raises:
            ClientError: when query or status requests fail. If the query was retried, the error
                         of the last try will be raised. The same exception is also captured in
                         ``query.error``. Raising can be prevented by setting ``raise_on_failure``
                         to ``False``.
            RunnerError: when a call to the query runner raises. This exception is raised
                         even if ``raise_on_failure` is ``False``, since it is likely an error
                         that is not just specific to this query.
            AlreadyRunningError: when another ``run_query()`` call on this query has not finished
                                 yet. This is not affected by ``raise_on_failure``.
        """
        if query.done:
            return  # nothing to do

        if not query._run_lock.acquire(blocking=False):
            raise AlreadyRunningError(kwargs=query.kwargs)

        try:
            if query.nb_tries > 0:
                query.reset()  # reset failed queries

            # query runner is invoked before every try, and once after the last try
            while True:
                await self._invoke_runner(query, raise_on_failure=raise_on_failure)
                if query.done:
                    return
                await self._try_query_once(query)
        finally:
            query._run_lock.release()

    async def _invoke_runner(self, query: Query, *, raise_on_failure: bool) -> None:
        """
        Invoke the query runner.

        Raises:
            ClientError: if the runner raises ``query.error``
            ValueError: if the runner raises a different ``ClientError`` than ``query.error``
            RunnerError: if the runner raises any other exception (which it shouldn't)
        """
        try:
            await self._runner(query)
        except ClientError as err:
            if err is not query.error:
                msg = "query runner raised a ClientError other than 'query.error'"
                raise ValueError(msg) from err
            if raise_on_failure:
                raise
        except AssertionError:
            raise
        except BaseException as err:
            raise RunnerError(cause=err) from err

    async def _try_query_once(self, query: Query) -> None:
        """A single iteration of running a query."""
        query._begin_try()

        try:
            await self._cooldown(query)

            req_timeout = _next_query_req_timeout(query)

            # pick the timeout we will use for the next try
            # TODO: not sure if this should also update the timeout setting in the Query state;
            #   for now, pass it as parameter to the _code() function
            next_timeout_secs = _next_timeout_secs(query)

            data = query._code(next_timeout_secs)

            query._begin_request()

            query.logger.info(f"call api for {query}")

            async with (
                _map_request_error(req_timeout),
                self._session().post(
                    url=urljoin(self._url, "interpreter"),
                    data=data,
                    timeout=req_timeout,
                ) as response,
            ):
                query._succeed_try(
                    response=await _result_or_raise(response, query.kwargs, query.logger),
                    response_bytes=response.content.total_bytes,
                )

        except CallTimeoutError as err:
            fail_with: ClientError = err
            if query.run_timeout_elapsed:
                assert query.run_duration_secs is not None
                fail_with = GiveupError(
                    cause=GiveupCause.RUN_TIMEOUT_DURING_QUERY_CALL,
                    kwargs=query.kwargs,
                    after_secs=query.run_duration_secs,
                )
            query._fail_try(fail_with)

        except ClientError as err:
            query._fail_try(err)

        finally:
            query._end_try()

    async def _cooldown(self, query: Query) -> None:
        """
        If the given query failed with ``TOO_MANY_QUERIES``, check for a cooldown period.

        Raises:
            ClientError: if the status request to find out the cooldown period fails
            GiveupError: if the cooldown is longer than the remaining run duration
        """
        logger = query.logger

        if not is_too_many_queries(query.error):
            return

        # If this client is running too many queries, we can check the status for a
        # cooldown period. This request failing is a bit of an edge case.
        # 'query.error' will be overwritten, which means we will not check for a
        # cooldown in the next iteration.
        status = await self._status(timeout=self._next_status_req_timeout(query))

        if not status.cooldown_secs:
            return

        run_duration = query.run_duration_secs
        assert run_duration is not None

        if run_timeout := query.run_timeout_secs:
            remaining = run_timeout - run_duration

            if status.cooldown_secs > remaining:
                logger.error(f"give up on {query} due to {status.cooldown_secs:.1f}s cooldown")
                raise GiveupError(
                    cause=GiveupCause.RUN_TIMEOUT_BY_COOLDOWN,
                    kwargs=query.kwargs,
                    after_secs=run_duration,
                )

        logger.info(f"{query} has cooldown for {status.cooldown_secs:.1f}s")
        await sleep(status.cooldown_secs)

    def _next_status_req_timeout(self, query: Query) -> aiohttp.ClientTimeout:
        """Status request timeout; possibly limited by either the run or status timeout settings."""
        remaining = None

        run_duration = query.run_duration_secs
        assert run_duration is not None

        if run_timeout := query.run_timeout_secs:
            remaining = run_timeout - run_duration

            if remaining <= 0.0:
                raise GiveupError(
                    cause=GiveupCause.RUN_TIMEOUT_BEFORE_STATUS_CALL,
                    kwargs=query.kwargs,
                    after_secs=run_duration,
                )

            if self._status_timeout_secs:
                remaining = min(remaining, self._status_timeout_secs)  # cap timeout if configured

        return aiohttp.ClientTimeout(total=remaining)


def _next_query_req_timeout(query: Query) -> aiohttp.ClientTimeout:
    """Query request timeout; possibly limited by either the run or request timeout settings."""
    run_total = None  # time left until "run_timeout_secs" exceeded
    query_total = None  # "[timeout:*]" setting plus "total_without_query_secs"

    run_duration = query.run_duration_secs
    assert run_duration is not None

    if run_timeout := query.run_timeout_secs:
        run_total = run_timeout - run_duration
        if run_total <= 0.0:
            raise GiveupError(
                cause=GiveupCause.RUN_TIMEOUT_BEFORE_QUERY_CALL,
                kwargs=query.kwargs,
                after_secs=run_duration,
            )

    if add_to_timeout_secs := query.request_timeout.total_without_query_secs:
        query_total = float(query.timeout_secs) + add_to_timeout_secs
        assert query_total > 0.0

    total = min(run_total, query_total) if run_total and query_total else run_total or query_total

    return aiohttp.ClientTimeout(
        total=total,
        connect=None,
        sock_connect=query.request_timeout.sock_connect_secs,
        sock_read=query.request_timeout.each_sock_read_secs,
    )


async def _parse_status(response: aiohttp.ClientResponse) -> Status:
    """Parses an /api/status response."""
    text = await response.text()

    slots: int | None = 0
    free_slots = None
    cooldown_secs = 0
    endpoint = None
    nb_running_queries = 0

    match_slots_overall = re.findall(r"Rate limit: (\d+)", text)
    match_slots_available = re.findall(r"(\d+) slots available now", text)
    match_cooldowns = re.findall(r"Slot available after: .+, in (\d+) seconds", text)
    match_endpoint = re.findall(r"Announced endpoint: (.+)", text)
    match_running_queries = re.findall(r"\d+\t\d+\t\d+\t\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", text)

    try:
        (slots_str,) = match_slots_overall
        slots = int(slots_str) or None

        endpoint = match_endpoint[0] if match_endpoint else None
        endpoint = None if endpoint == "none" else endpoint

        nb_running_queries = len(match_running_queries)

        if slots:
            cooldowns = [int(secs) for secs in match_cooldowns]

            if match_slots_available:
                free_slots_str = match_slots_available[0]
                free_slots = int(free_slots_str)
            else:
                free_slots = slots - len(cooldowns)

            cooldown_secs = 0 if free_slots > 0 else min(cooldowns)
    except ValueError as err:
        await _raise_for_response(response, cause=err)

    return Status(
        slots=slots,
        free_slots=free_slots,
        cooldown_secs=cooldown_secs,
        endpoint=endpoint,
        nb_running_queries=nb_running_queries,
    )


@asynccontextmanager
async def _map_request_error(
    timeout: ClientTimeout | None = None,
) -> AsyncIterator[None]:
    """Context to make requests in; maps errors to our exception types."""
    try:
        yield
    except aiohttp.ClientError as err:
        await _raise_for_request_error(err)
    except TimeoutError as err:
        assert timeout is not None
        assert timeout.total
        raise CallTimeoutError(cause=err, after_secs=timeout.total) from err


def _next_timeout_secs(query: Query) -> int:
    """
    Decide the ``[timeout:*]`` setting for the given query's next try.

     - pick ``Query.timeout_secs``
     - cap it by ``Query.run_timeout_secs``, and the time that has elapsed so far
     - if a previous try was cancelled with less or equal than the chosen time,
       raise without trying again

    Raises:
        GiveupError: when the run timeout elapsed, or when a previous try timed out with
                     a lower timeout setting
    """
    timeout_secs = query.timeout_secs

    if (at_most := query._run_duration_left_secs) and at_most < timeout_secs:
        query._logger.info(f"{query} has {at_most}s left to run")
        at_most = math.floor(at_most)
        query._logger.info(f"adjust {query} [timeout:{timeout_secs}] to [timeout:{at_most}]")
        timeout_secs = at_most
    else:
        query._logger.info(f"{query} will use [timeout:{timeout_secs}]")

    not_enough = query._max_timed_out_after_secs or 0

    if timeout_secs <= not_enough:
        query._logger.error(f"{query} previously timed out with [timeout:{not_enough}] - give up")
        raise GiveupError(
            cause=GiveupCause.EXPECTING_QUERY_TIMEOUT,
            kwargs=query.kwargs,
            after_secs=query.run_duration_secs or 0.0,
        )

    return timeout_secs



import asyncio
import math
from dataclasses import dataclass

from aio_overpass._env import IS_UNIT_TEST


__docformat__ = "google"
__all__ = (
    "sleep",
    "time",
    "Instant",
)


class _ClockMock:
    """Mock alternative to event loop time."""

    __slots__ = ("_time",)

    def __init__(self) -> None:
        self._time = 0.0

    async def mock_sleep(self, delay: float) -> None:
        self._time += delay
        await asyncio.sleep(0.0)

    def mock_time(self) -> float:
        return self._time


if IS_UNIT_TEST:
    _clock_mock = _ClockMock()
    sleep = _clock_mock.mock_sleep
    time = _clock_mock.mock_time
else:
    sleep = asyncio.sleep
    time = asyncio.get_event_loop().time


@dataclass(kw_only=True, slots=True, frozen=True, repr=False, order=True)
class Instant:
    """
    Measurement of a monotonic clock.

    Attributes:
        when: the current time, according to the event loop's internal monotonic clock
              (details are unspecified and may differ per event loop).
    """

    when: float

    @classmethod
    def now(cls) -> "Instant":
        return cls(when=time())

    @property
    def ceil(self) -> int:
        return math.ceil(self.when)

    @property
    def elapsed_secs_since(self) -> float:
        return time() - self.when

    def __sub__(self, earlier: "Instant") -> float:
        if self.when < earlier.when:
            msg = f"{self} is earlier than {earlier}"
            raise ValueError(msg)
        return self.when - earlier.when

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.when:.02f})"
